import numpy as np
import warnings
import os
import signal
import subprocess
import pprint
from .video_info import run_async, get_info, get_num_NVIDIA_GPUs, decoder_to_nvidia


class FFmpegReader:
    def __repr__(self):
        props = pprint.pformat(self.__dict__).replace('{',' ').replace('}',' ')
        return f'{self.__class__}\n'  + props
    
    def __len__(self):
        return self.count

    def __iter__(self):
        return self

    def __next__(self):
        ret, img = self.read()
        if ret:
            return img
        else:
            raise StopIteration

    @staticmethod
    def VideoReader(filename, codec, pix_fmt, 
                    resize, resize_keepratio, resize_keepratioalign):
        assert pix_fmt in ['rgb24', 'bgr24']

        vid = FFmpegReader()
        videoinfo = get_info(filename)
        vid.width = videoinfo.width
        vid.height = videoinfo.height
        vid.fps = fps = videoinfo.fps
        vid.count = videoinfo.count
        vid.origin_width, vid.origin_height = vid.width, vid.height
        vid.crop_width, vid.crop_height = vid.width, vid.height
        
        codecopt = '-c:v ' + codec if codec else ''

        if resize and resize!=(vid.crop_width, vid.crop_height):
            vid.width, vid.height = dst_width, dst_height = resize
            if resize_keepratio:
                re_width, re_height = vid.origin_width/(vid.origin_height / dst_height) , dst_height
                if re_width > dst_width:
                    re_width, re_height = dst_width, vid.origin_height/(vid.origin_width / dst_width)
                re_width, re_height = int(re_width), int(re_height)
                scaleopt = '-vf scale=%d:%d' % (re_width, re_height)
                if resize_keepratioalign is None: resize_keepratioalign = 'center'
                paddings = {'center': ((dst_width - re_width) // 2, (dst_height - re_height) // 2),
                            'topleft': (0, 0),
                            'topright': (dst_width - re_width, 0),
                            'bottomleft': (0, dst_height - re_height), 
                            'bottomright': (dst_width - re_width, dst_height - re_height)}
                assert resize_keepratioalign in paddings, 'resize_keepratioalign must be one of "center"(mmpose), "topleft"(mmdetection), "topright", "bottomleft", "bottomright"'
                xpading, ypading = paddings[resize_keepratioalign]
                padopt = f'pad={dst_width}:{dst_height}:{xpading}:{ypading}:black'
                filteropt = f'{scaleopt},{padopt}'
            else:
                filteropt = '-vf scale=%d:%d' % (dst_width, dst_height)
        else:
            filteropt = ''

        vid.size = (vid.width, vid.height)
        args = (f'ffmpeg -loglevel warning {codecopt} -r {fps} -i "{filename}" '
                f'{filteropt} -pix_fmt {pix_fmt} '
                f'-r {fps} -f rawvideo pipe:')
        vid.process = run_async(args)
        return vid

    def read(self):
        in_bytes = self.process.stdout.read(self.height * self.width * 3)
        if not in_bytes:
            return False, None
        img = None
        img = np.frombuffer(in_bytes, np.uint8).reshape([self.height, self.width, 3])
        return True, img

    def release(self):
        os.killpg(os.getpgid(self.process.pid), signal.SIGKILL)
        self.process.terminate()
        self.process.wait()


class FFmpegReaderNV(FFmpegReader):
    @staticmethod
    def VideoReader(filename, pix_fmt, crop_xywh, 
                    resize, resize_keepratio, resize_keepratioalign, 
                    gpu):
        assert pix_fmt in ['rgb24', 'bgr24']
        numGPU = get_num_NVIDIA_GPUs()
        assert numGPU>0, 'No GPU found'
        gpu = int(gpu) % numGPU if gpu is not None else 0
        assert resize is None or len(resize) == 2, 'resize must be a tuple of (width, height)'
        videoinfo = get_info(filename)
        vid = FFmpegReaderNV()
        vid.origin_width = videoinfo.width
        vid.origin_height = videoinfo.height
        vid.fps = videoinfo.fps
        vid.count = videoinfo.count
        vid.width, vid.height = vid.origin_width, vid.origin_height
        vid.codec = videoinfo.codec
        vid.codecNV = decoder_to_nvidia(vid.codec)
        
        if crop_xywh:
            crop_w, crop_h = crop_xywh[2:]
            vid.width, vid.height = crop_w, crop_h
            x, y, w, h = crop_xywh
            top, bottom, left, right = y, vid.origin_height - (y + h), x, vid.origin_width - (x + w)  #crop length
            cropopt = f'-crop {top}x{bottom}x{left}x{right}'
        else:
            crop_w, crop_h = vid.origin_width, vid.origin_height
            cropopt = ''

        vid.crop_width, vid.crop_height = crop_w, crop_h

        if resize and resize!=(vid.crop_width, vid.crop_height):
            vid.width, vid.height = dst_width, dst_height = resize
            if resize_keepratio:
                re_width, re_height = crop_w/(crop_h / dst_height) , dst_height
                if re_width > dst_width:
                    re_width, re_height = dst_width, crop_h/(crop_w / dst_width)
                re_width, re_height = int(re_width), int(re_height)
                scaleopt = f'-vf scale_cuda={re_width}:{re_height},hwdownload,format=nv12'
                if resize_keepratioalign is None: resize_keepratioalign = 'center'
                paddings = {'center': ((dst_width - re_width) // 2, (dst_height - re_height) // 2),
                            'topleft': (0, 0),
                            'topright': (dst_width - re_width, 0),
                            'bottomleft': (0, dst_height - re_height), 
                            'bottomright': (dst_width - re_width, dst_height - re_height)}
                assert resize_keepratioalign in paddings, 'resize_keepratioalign must be one of "center"(mmpose), "topleft"(mmdetection), "topright", "bottomleft", "bottomright"'
                xpading, ypading = paddings[resize_keepratioalign]
                padopt = f'pad={dst_width}:{dst_height}:{xpading}:{ypading}:black'
                filteropt = f'{scaleopt},{padopt}'
            else:
                filteropt = f'-vf scale_cuda={dst_width}:{dst_height},hwdownload,format=nv12'
        else:
            filteropt = '-vf hwdownload,format=nv12'
        
        args = (f'ffmpeg -loglevel warning -hwaccel cuda -hwaccel_device {gpu} -hwaccel_output_format cuda '
                f' -vcodec {vid.codecNV} {cropopt} -r {vid.fps} -i "{filename}" '
                f' {filteropt} -pix_fmt {pix_fmt} -r {vid.fps} -f rawvideo pipe:')

        vid.process = run_async(args)
        vid.size = (vid.width, vid.height)
        return vid



import numpy as np
import subprocess
import warnings
import pprint
from .video_info import run_async, release_process, get_num_NVIDIA_GPUs


class FFmpegWriter:
    def __del__(self):
        if hasattr(self, 'process'):
            self.release()

    def __repr__(self):
        props = pprint.pformat(self.__dict__).replace('{',' ').replace('}',' ')
        return f'{self.__class__}\n'  + props

    @staticmethod
    def VideoWriter(filename, codec, fps, frameSize, pix_fmt):
        if codec is None:
            codec = 'libx264'
        elif not isinstance(codec, str):
            codec = 'libx264'
            warnings.simplefilter('''
                Codec should be a string. Eg `h264`, `h264_nvenc`. 
                You may used CV2.VideoWriter_fourcc, which will be ignored.
                ''')
        
        vid = FFmpegWriter()
        vid.fps, vid.size = fps, frameSize
        vid.width, vid.height = vid.size if vid.size else (None, None)
        vid.codec, vid.pix_fmt, vid.filename = codec, pix_fmt, filename
        vid.waitInit = True
        return vid

    def _init_video_stream(self):
        args = (f'ffmpeg -y -loglevel warning ' 
                f'-f rawvideo -pix_fmt {self.pix_fmt} -s {self.width}x{self.height} -r {self.fps} -i pipe: '
                f'-r {self.fps} -c:v {self.codec} -pix_fmt yuv420p "{self.filename}"')
        self.process = run_async(args)

    def write(self, img):
        if self.waitInit:
            if self.size is None:
                self.size = (img.shape[1], img.shape[0])          
            self.width, self.height = self.size
            self._init_video_stream()
            self.waitInit = False

        assert self.size == (img.shape[1], img.shape[0])
        img = img.astype(np.uint8).tobytes()
        self.process.stdin.write(img)

    def release(self):
        self.process.terminate()
        self.process.wait()


class FFmpegWriterNV(FFmpegWriter):
    @staticmethod
    def VideoWriter(filename, codec, fps, frameSize, pix_fmt, gpu):
        numGPU = get_num_NVIDIA_GPUs()
        assert numGPU
        gpu = int(gpu) % numGPU if gpu is not None else 0
        if codec is None:
            codec = 'hevc_nvenc'
        elif not isinstance(codec, str):
            codec = 'hevc_nvenc'
            warnings.simplefilter('''
                Codec should be a string. Eg `h264`, `h264_nvenc`. 
                You may used CV2.VideoWriter_fourcc, which will be ignored.
                ''')
        elif codec.endswith('_nvenc'):
            codec = codec
        else:
            codec = codec + '_nvenc'
        assert codec in ['hevc_nvenc', 'h264_nvenc'], 'codec should be `hevc_nvenc` or `h264_nvenc`'

        vid = FFmpegWriterNV()
        vid.fps, vid.size = fps, frameSize
        vid.width, vid.height = vid.size if vid.size else (None, None)
        vid.codec, vid.pix_fmt, vid.filename = codec, pix_fmt, filename
        vid.gpu = gpu
        vid.waitInit = True
        return vid

    def _init_video_stream(self):
        args = (f'ffmpeg -y -loglevel warning '
            f'-f rawvideo -pix_fmt {self.pix_fmt} -s {self.width}x{self.height} -r {self.fps} -i pipe: '
            f'-r {self.fps} -gpu {self.gpu} -c:v {self.codec} -pix_fmt yuv420p "{self.filename}"')
        self.process = run_async(args)


from .ffmpeg_reader import FFmpegReader, FFmpegReaderNV
from .ffmpeg_writer import FFmpegWriter, FFmpegWriterNV
from .video_info import get_num_NVIDIA_GPUs
import shutil
from subprocess import Popen, PIPE


def _check():
    if not shutil.which('ffmpeg'):
        raise RuntimeError('ffmpeg is not installed')
    if not shutil.which('ffprobe'):
        raise RuntimeError('ffprobe is not installed')

_check()

_check_nvidia_init = None

def _check_nvidia():
    global _check_nvidia_init
    if _check_nvidia_init is None:
        cmd = 'ffmpeg'
        p = Popen([cmd], shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate(b"")
        stderr = stderr.decode('utf-8')
        p.terminate()

        if ('--enable-cuda-nvcc' not in stderr 
            or '--enable-cuvid' not in stderr
            or '--enable-nvenc' not in stderr):
            raise RuntimeError('The ffmpeg is not compiled with CUDA support')

        if get_num_NVIDIA_GPUs() == 0:
            raise RuntimeError('No NVIDIA GPU found')

        _check_nvidia_init = True

    return True


def VideoCapture(file, 
                 codec=None, 
                 pix_fmt='bgr24',
                 resize=None,
                 resize_keepratio=True,
                 resize_keepratioalign='center'):
    """
    Alternative to cv2.VideoCapture

    Parameters
    ----------
    file : str
        Path to video file.
    codec : str
        Codec to use. Optional. Default is `None`.
    pix_fmt : str
        Pixel format. ['bgr24' | 'rgb24']. Optional. Default is 'bgr24'.
    resize  : tuple
        Resize the video to the given size. Optional. Default is `None`.
    resize_keepratio : bool
        Keep the aspect ratio and the border is black. Optional. Default is `True`.
    resize_keepratioalign : str
        Align the image to the `center`, `topleft`, `topright`, `bottomleft` or
        `bottomright`. Optional. Default is 'center'.

    Examples
    --------
    opencv
    ```
    cap = cv2.VideoCapture(file)
    while True:
        ret, frame = cap.read()
        if not ret:
            break
        pass
    ```

    ffmpegcv
    ```
    cap = ffmpegcv.VideoCapture(file)
    while True:
        ret, frame = cap.read()
        if not ret:
            break
        pass
    ```

    Or use iterator
    ```
    cap = ffmpegcv.VideoCapture(file)
    for frame in cap:
        pass
    counts = len(cap)
    ```

    Use GPU to accelerate decoding
    ```
    cap_cpu = ffmpegcv.VideoCapture(file)
    cap_gpu = ffmpegcv.VideoCaptureNV(file)
    ```

    Use rgb24 instead of bgr24
    ```
    cap = ffmpegcv.VideoCapture(file, pix_fmt='rgb24')
    ```

    Resize the video to the given size
    ```
    cap = ffmpegcv.VideoCapture(file, resize=(640, 480))
    ```

    Resize and keep the aspect ratio with black border
    ```
    cap = ffmpegcv.VideoCapture(file, resize=(640, 480), resize_keepratio=True)
    ```

    Author: Chenxinfeng 2022-04-16, cxf529125853@163.com
    """
    print(file, codec, pix_fmt, resize, resize_keepratio)
    return FFmpegReader.VideoReader(file, codec, pix_fmt, 
                                    resize, resize_keepratio, resize_keepratioalign)


def VideoReader(*args, **kwargs):
    """
     `ffmpegcv.VideoReader` is an alias to `ffmpegcv.VideoCapture`
    """
    return VideoCapture(*args, **kwargs)                                                                       


def VideoWriter(file, 
                codec=None, 
                fps=30, 
                frameSize=None, 
                pix_fmt='bgr24'):
    """
    Alternative to cv2.VideoWriter

    Parameters
    ----------
    file : str
        Path to video file.
    codec : str
        Codec to use. Optional. Default is `None` (x264).
    fps : number
        Frames per second. Optional. Default is 30.
    frameSize : tuple
        Frame size. (width, height). Optional. Default is `None`, which is 
        decided by the size of the first frame.
    pix_fmt : str
        Pixel format of input. ['bgr24' | 'rgb24']. Optional. Default is 'bgr24'.

    Examples
    --------
    opencv
    ```
    out = cv2.VideoWriter('outpy.avi',
                          cv2.VideoWriter_fourcc('M','J','P','G'), 
                          10, 
                          (w, h))
    out.write(frame1)
    out.write(frame2)
    out.release()
    ```

    ffmpegcv
    ```
    out = ffmpegcv.VideoWriter('outpy.avi', None, 10, (w, h))
    out.write(frame1)
    out.write(frame2)
    out.release()
    ```

    frameSize is decided by the size of the first frame
    ```
    out = ffmpegcv.VideoWriter('outpy.avi', None, 10)
    ```

    Use GPU to accelerate encoding
    ```
    out_cpu = ffmpegcv.VideoWriter('outpy.avi', None, 10)
    out_gpu = ffmpegcv.VideoWriter('outpy.avi', 'h264_nvenc', 10)
    ```

    Use rgb24 instead of bgr24
    ```
    out = ffmpegcv.VideoWriter('outpy.avi', None, 10, pix_fmt='rgb24')
    out.write(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
    ```

    Author: Chenxinfeng 2022-04-16, cxf529125853@163.com
    """
    return FFmpegWriter.VideoWriter(file, codec, fps, frameSize, pix_fmt)


def VideoCaptureNV(file,
                   pix_fmt='bgr24',
                   crop_xywh=None,
                   resize=None,
                   resize_keepratio=True,
                   resize_keepratioalign='center',
                   gpu=0):
    """
    `ffmpegcv.VideoCaptureNV` is a gpu version for `ffmpegcv.VideoCapture`.
    """
    _check_nvidia()
    return FFmpegReaderNV.VideoReader(file, pix_fmt, crop_xywh, 
                                      resize, resize_keepratio, resize_keepratioalign, 
                                      gpu)


def VideoReaderNV(*args, **kwargs):
    """
     `ffmpegcv.VideoReaderNV` is an alias to `ffmpegcv.VideoCaptureNV`
    """
    return VideoCaptureNV(*args, **kwargs)


def VideoWriterNV(file,
                  codec=None,
                  fps=30,
                  frameSize=None,
                  pix_fmt='bgr24',
                  gpu=0):
    """
    `ffmpegcv.VideoWriterNV` is a gpu version for `ffmpegcv.VideoWriter`.
    """
    _check_nvidia()
    return FFmpegWriterNV.VideoWriter(file, codec, fps, frameSize, pix_fmt, gpu)



# %%
import os
import signal
import subprocess
from subprocess import Popen, PIPE
import re
from collections import namedtuple
import xml.etree.ElementTree as ET


def get_info(video):
    cmd = 'ffprobe -v quiet -print_format xml -select_streams v:0 -show_format -show_streams "{}"'.format(video)
    output = subprocess.check_output(cmd, shell=True)
    root = ET.fromstring(output)
    assert (root[0].tag, root[0][0].tag) == ('streams', 'stream')
    vinfo = root[0][0].attrib

    VideoInfo = namedtuple('VideoInfo', ['width', 'height', 'fps', 'count', 
                                        'codec', 'duration'])
    outinfo = dict()
    outinfo['width'] = int(vinfo['width'])
    outinfo['height'] = int(vinfo['height'])
    outinfo['fps'] = eval(vinfo['r_frame_rate'])
    outinfo['count'] = int(vinfo['nb_frames'])
    outinfo['codec'] = vinfo['codec_name']
    outinfo['duration'] = float(vinfo['duration'])
    videoinfo = VideoInfo(**outinfo)

    return videoinfo


def get_num_NVIDIA_GPUs():
    cmd = 'ffmpeg -f lavfi -i nullsrc -c:v h264_nvenc -gpu list -f null -'
    p = Popen([cmd], shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate(b"")
    p.terminate()
    pattern = re.compile(r'GPU #\d+ - < NVIDIA')
    nv_info = pattern.findall(stderr.decode())
    num_gpu = len(nv_info)
    return num_gpu


def encoder_to_nvidia(codec):
    codec_map = {'h264'      : 'h264_nvenc',
                 'hevc'      : 'hevc_nvenc'}

    if codec in codec_map:
        return codec_map[codec]
    elif codec in codec_map.values():
        return codec
    else:
        raise Exception('No NV codec found for %s' % codec)
    

def decoder_to_nvidia(codec):
    codec_map = {'av1'       : 'av1_cuvid',
                 'h264'      : 'h264_cuvid',
                 'x264'      : 'h264_cuvid',    
                 'hevc'      : 'hevc_cuvid',
                 'x265'      : 'hevc_cuvid',
                 'h265'      : 'hevc_cuvid',
                 'mjpeg'     : 'mjpeg_cuvid',
                 'mpeg1video': 'mpeg1_cuvid',
                 'mpeg2video': 'mpeg2_cuvid',
                 'mpeg4'     : 'mpeg4_cuvid',
                 'vp1'       : 'vp1_cuvid',
                 'vp8'       : 'vp8_cuvid',
                 'vp9'       : 'vp9_cuvid'}
    
    if codec in codec_map:
        return codec_map[codec]
    elif codec in codec_map.values():
        return codec
    else:
        raise Exception('No NV codec found for %s' % codec)


def run_async(args):
    quiet = True
    stderr_stream = subprocess.DEVNULL if quiet else None
    return subprocess.Popen(
        args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=stderr_stream, 
        shell=True, preexec_fn=os.setsid
    )


def release_process(process):
    try:
        os.killpg(os.getpgid(process.pid), signal.SIGKILL)
    except:
        pass
    process.terminate()
    process.wait()



#setup.py
from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='ffmpegcv', # 应用名
    version='0.1.6', # 版本号
    packages=find_packages(include=['ffmpegcv*']), # 包括在安装包内的 Python 包
    author='chenxf',
    author_email='cxf529125853@163.com',
    url='https://pypi.org/project/ffmpegcv/',
    long_description=long_description,
    long_description_content_type='text/markdown'
)
